import hashlib
from typing import Any

from baml_agents._baml_client_proxy._hooks._implementations._test_generator_helpers import (
    get_args_block_str,
)
from baml_agents._baml_client_proxy._hooks._on_before_call_hook import (
    OnBeforeCallHookContext,
    OnBeforeCallHookSync,
)


def generate_baml_test(test_name, baml_function_name, params):
    return f"""\
test {test_name} {{
  functions [{baml_function_name}]
  args {{
{get_args_block_str(params)}
  }}
}}\n\n"""


def get_test_name(params, baml_function_name):
    hash_input = f"{baml_function_name}{params}"
    return f"t_{hashlib.md5(hash_input.encode()).hexdigest()[:6]}"  # noqa: S324


class BamlTestGeneratorHook(OnBeforeCallHookSync):
    def __init__(self):
        self.baml_test_source_code = ""

    def on_before_call(self, params: dict[str, Any], ctx: OnBeforeCallHookContext):
        test_name = get_test_name(params, ctx.baml_function_name)
        self.baml_test_source_code += generate_baml_test(
            test_name, ctx.baml_function_name, params
        )



import asyncio
import html
import inspect
import json
import uuid
from collections.abc import Callable
from pprint import pformat
from typing import Any, Generic, Literal, TypeVar, cast

from baml_py import Collector
from IPython.display import Javascript, display

from baml_agents._utils._sole import sole

from ._jupyter_button_hide import (
    hide_text_under_a_button,
    hide_text_under_a_button_nested,
)
from ._token_cost_estimator import TokenCostEstimator

T = TypeVar("T")


def _format_price_and_duration(
    cost_usd_per_thousand: float, duration_ms: int | None
) -> str:
    duration_sec = f"{duration_ms/1000:.2f}s" if duration_ms is not None else "N/A"
    return f"{cost_usd_per_thousand:.2f}$/1k | {duration_sec}"


def _get_call_duration_ms(call) -> int | None:
    timing = getattr(call, "timing", None)
    if timing is not None:
        return getattr(timing, "duration_ms", None)
    return None


def _get_log_duration_ms(log) -> int | None:
    timing = getattr(log, "timing", None)
    if timing is not None:
        return getattr(timing, "duration_ms", None)
    return None


class JupyterOutputBox:
    """
    Context manager that:
      - Renders pretty <pre> formatted output (preserves whitespace and newlines).
      - Updates the content efficiently via Jupyter's display_id mechanism.
      - Removes the HTML element on exit, hiding the output when done.
    """

    def __init__(
        self,
        initial_message: str = "Initializing…",
        *,
        display_id: str | None = None,
        clear_after_finish: bool = True,
    ):
        self.display_id = display_id or f"stream-{uuid.uuid4()}"
        self.clear_after_finish = clear_after_finish
        escaped_msg = html.escape(initial_message, quote=False)
        self._initial_html = f'<pre id="{self.display_id}">{escaped_msg}</pre>'
        self._active = False
        self._handle = None

    def __enter__(self):
        if not self._active:
            self._handle = display(
                {"text/html": self._initial_html},
                display_id=self.display_id,
                raw=True,
            )
            self._active = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.clear_after_finish and self._active:
            js = f"""
            (function(){{
              var el = document.getElementById("{self.display_id}");
              if (el) el.remove();
            }})();
            """
            display(Javascript(js))
        self._active = False
        self._handle = None

    def update(self, data, *, use_br: bool = False):
        if not isinstance(data, str):
            try:
                s = data.model_dump_json(indent=4)
            except AttributeError:
                s = json.dumps(data, indent=4)
        else:
            s = data

        escaped = html.escape(s, quote=False)

        if use_br:
            html_body = escaped.replace("\n", "<br>")
            new_html = (
                f'<div id="{self.display_id}" style="font-family:monospace">'
                f"{html_body}</div>"
            )
        else:
            new_html = f'<pre id="{self.display_id}">{escaped}</pre>'

        if self._handle is None:
            raise RuntimeError(
                "Display handle is None. Did you forget to enter the context manager?"
            )
        self._handle.update({"text/html": new_html}, raw=True)

    def display(
        self,
        formatter: Literal["json", "pformat"] | Callable[[Any], None] | None = None,
    ):
        formatter = formatter or "pformat"

        def update_result(result):
            if formatter == "json":
                try:
                    s = result.model_dump_json(indent=4)
                except AttributeError:
                    s = json.dumps(result, indent=4)
                self.update(s)
            elif formatter == "pformat":
                self.update(pformat(result.model_dump(), width=200, sort_dicts=False))
            elif callable(formatter):
                formatter(result)
            else:
                raise ValueError(f"Unknown formatter: {formatter}")

        return update_result


class _StreamingInterceptorWrapper:
    def __init__(self, ai: Any, callback: Callable, /):
        self._ai = ai
        self._callback = callback

    def __getattribute__(self, name: str) -> Any:
        # Use object.__getattribute__ to access instance variables directly
        # to avoid recursive calls to this __getattribute__ method.
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            # If the attribute is not directly on the wrapper, get it from the wrapped object.
            pass

        ai_instance = object.__getattribute__(self, "_ai")
        callback = object.__getattribute__(self, "_callback")

        # Special handling for methods required by the wrapper itself or python internals
        # Note: __getattr__ might be simpler if we only intercept non-dunder methods.
        # But __getattribute__ is more powerful for intercepting *everything*.
        if name in {
            "_ai",
            "_callback",
            "__class__",
            "__init__",
            "__getattribute__",
            "__setattr__",  # Added for completeness, though not strictly needed for the example
            "__delattr__",  # Added for completeness
            "__dict__",
            "__dir__",  # Often useful to proxy dir() calls
            "__repr__",  # Might want to customize repr
            "__str__",  # Might want to customize str
            # Async specific methods often checked by libraries
            "__await__",
            "__aiter__",
            "__anext__",
            "__aenter__",
            "__aexit__",
        }:
            # If it's one of our internal/special methods, get it directly from self
            # This was handled by the try/except block above now. Redundant but kept for clarity.
            return object.__getattribute__(self, name)

        # Get the attribute from the *stream* property of the wrapped object
        # This is specific to the Streaming Interceptor's logic
        attr = getattr(ai_instance.stream, name)

        # Only wrap if it's callable (likely a method)
        if not callable(attr):
            return attr  # Return non-callable attributes directly

        # --- Streaming Specific Wrapping Logic ---
        # This wrapper assumes the underlying method is async and returns a stream
        async def stream_wrapper(*args, **kwargs):
            stream_obj = attr(*args, **kwargs)
            async for partial in stream_obj:
                callback(partial)
            # Ensure get_final_response exists and is awaitable if needed
            # This part might need adjustment based on the actual stream object interface
            final_response = None
            if hasattr(
                stream_obj, "get_final_response"
            ) and inspect.iscoroutinefunction(stream_obj.get_final_response):
                final_response = await stream_obj.get_final_response()
                callback(final_response)
            # What if get_final_response doesn't exist or isn't async? Need fallback?
            # For now, assume it exists and is async as per the original code.
            return (
                final_response  # Or maybe the stream_obj itself if no final response?
            )

        # Copy metadata
        stream_wrapper.__name__ = name
        stream_wrapper.__qualname__ = f"{type(self).__name__}.{name}"
        stream_wrapper.__doc__ = getattr(attr, "__doc__", None)
        stream_wrapper.__annotations__ = getattr(attr, "__annotations__", {})
        # Consider adding functools.wraps for more robust metadata copying

        return stream_wrapper


class JupyterBamlCollector(Generic[T]):
    def __init__(
        self,
        ai: T,
        *,
        stream_callback: Callable | None = None,
        intent_summarizer: Callable | None = None,
    ):
        self._original_ai = ai
        self.collector = Collector(name="collector")
        self._ai = self._original_ai.with_options(collector=self.collector)  # type: ignore
        self.cost_estimator = TokenCostEstimator()
        self._stream_callback = stream_callback
        self._intent_summarizer = intent_summarizer

    @property
    def ai(self) -> T:
        if self._stream_callback:
            return _StreamingInterceptorWrapper(self._ai, self._stream_callback)  # type: ignore
        return self._ai

    @property
    def b(self) -> T:
        return self.ai

    @staticmethod
    def _format_log_messages(messages):
        prompt_parts = []
        for message in messages:
            content = sole(message["content"])
            if content["type"] != "text":
                raise ValueError(
                    f"Expected content type 'text', but got '{content['type']}'"
                )
            prompt_parts.append(f"[{message['role']}]\n{content['text']}")
        return "\n\n".join(prompt_parts)

    def _get_prompt_buttons(self, calls, *, omit_cost_and_model: bool = False):
        ret = {}
        for call in calls:
            request_body = call.http_request.body.json()
            messages = request_body["messages"]

            cost = self.cost_estimator.calculate_cost(
                request_body["model"],
                call.usage.input_tokens,
                call.usage.output_tokens,
            )
            cost_usd_per_thousand = cost["total_cost_usd"] * 1_000
            duration_ms = _get_call_duration_ms(call)

            formatted = self._format_log_messages(messages)
            if omit_cost_and_model:
                label = f"Prompt | messages={len(messages)} |"
            else:
                price_and_duration = _format_price_and_duration(
                    cost_usd_per_thousand, duration_ms
                )
                label = (
                    f"Prompt | messages={len(messages)} | "
                    f"{price_and_duration} | {cost['model_info']['model_name']}"
                )
            ret[label] = formatted
        return ret

    async def _get_completion_button(
        self, raw_llm_response, log=None, *, omit_cost_and_model: bool = False
    ):
        summary_label = None
        if self._intent_summarizer is not None:
            summary_label = await self._intent_summarizer(raw_llm_response)
            # Defensive: fallback to default label if summarizer fails
            summary_label = None

        price_and_duration = ""
        model_name = ""
        if not omit_cost_and_model and log is not None and log.calls:
            last_call = log.calls[-1]
            request_body = last_call.http_request.body.json()
            cost = self.cost_estimator.calculate_cost(
                request_body["model"],
                last_call.usage.input_tokens,
                last_call.usage.output_tokens,
            )
            cost_usd_per_thousand = cost["total_cost_usd"] * 1_000
            duration_ms = _get_call_duration_ms(last_call)
            price_and_duration = (
                f" { _format_price_and_duration(cost_usd_per_thousand, duration_ms) }"
            )
            model_name = f" | {cost['model_info']['model_name']}"
        # Use summarizer output as label if available, else fallback to default
        if summary_label is not None:
            label = f"Completion | {summary_label}"
        else:
            suffix = ""
            if not omit_cost_and_model:
                suffix = f"{price_and_duration}{model_name}"
            label = f"Completion{' |'if suffix else ''}{suffix}"
        return label, str(raw_llm_response)

    async def display_calls(
        self,
        *,
        prompts: Literal["always_hide", "always_show", "show", "hide"] = "hide",
        completions: Literal["always_hide", "always_show", "show", "hide"] = "hide",
    ):
        logs = list(self.collector.logs)
        show_step = len(logs) > 1

        def get_key(i, k):
            if show_step:
                return f"{i} | {k}"
            return k

        completion_button_coros = [
            (
                self._get_completion_button(log.raw_llm_response, log=log)
                if completions != "always_hide"
                else None
            )
            for log in logs
        ]
        completion_buttons = (
            await asyncio.gather(
                *[coro for coro in completion_button_coros if coro is not None]
            )
            if completions != "always_hide"
            else []
        )

        completion_idx = 0
        for i, log in enumerate(logs, start=1):
            if prompts != "always_hide":
                for k, v in self._get_prompt_buttons(log.calls).items():
                    hide_text_under_a_button(get_key(i, k), v, visibility=prompts)
            if completions != "always_hide":
                k, v = completion_buttons[completion_idx]
                hide_text_under_a_button(get_key(i, k), v, visibility=completions)
                completion_idx += 1

    async def display_session(
        self,
        root_name: str,
        *,
        show_depth=0,
        prompts: Literal["always_hide", "hide"] = "hide",
        completions: Literal["always_hide", "hide"] = "hide",
        show_everything: bool = False,
    ):
        nested_buttons = {}
        logs = list(self.collector.logs)  # Ensure it's a list for indexing

        # --- Populate nested_buttons with unique keys ---
        # Gather completion buttons first as they might involve async calls (summarizer)
        completion_button_coros = []
        if completions != "always_hide":
            completion_button_coros = [
                self._get_completion_button(
                    log.raw_llm_response,
                    log=log,
                    omit_cost_and_model=False,  # Keep True as per original logic for this func
                )
                for log in logs
            ]

        # Gather results, handling potential exceptions during gather
        completion_results = []
        if completions != "always_hide":
            completion_results = await asyncio.gather(
                *completion_button_coros, return_exceptions=True
            )

        # Now iterate through logs *with their index* to build the final dict
        for i, log in enumerate(logs, start=1):
            log_prefix = f"{i} | "

            # 1. Add Prompt Buttons for this log with unique keys
            # Use omit_cost_and_model=False to show cost/model details on prompts here
            prompt_buttons = self._get_prompt_buttons(
                log.calls, omit_cost_and_model=False
            )
            for k, v in prompt_buttons.items():
                # Add index prefix for guaranteed uniqueness
                unique_key = f"{log_prefix}{k}"
                if prompts != "always_hide":
                    nested_buttons[unique_key] = v

            # 2. Add Completion Button for this log with unique key
            if completions != "always_hide":
                completion_data = completion_results[i - 1]  # Use i-1 for 0-based index

                if isinstance(completion_data, Exception):
                    raise completion_data

                if completion_data is not None:  # Check if button generation succeeded
                    completion_k, completion_v = completion_data  # type: ignore
                    # Add index prefix for guaranteed uniqueness
                    unique_key = f"{log_prefix}{completion_k}"
                    nested_buttons[unique_key] = completion_v
                else:
                    # Handle case where button generation returned None unexpectedly
                    nested_buttons[f"{log_prefix}Completion (Not Available)"] = (
                        "Completion data could not be generated."
                    )

        # --- Display using the populated nested_buttons ---
        root_label = f"{root_name} | Cost: {self.format_total_cost()}"
        hide_text_under_a_button_nested(
            root_label,
            nested_buttons,
            visibility="always_show" if show_everything else "hide",
            hide_after_level=show_depth,
        )

    def format_total_cost(self):
        all_models = set()
        total_cost_usd_per_thousand = 0
        total_duration_ms = 0
        for log in self.collector.logs:
            log_duration_ms = _get_log_duration_ms(log)
            if log_duration_ms is not None:
                total_duration_ms += log_duration_ms
            for call in log.calls:
                if call.http_request is None:
                    raise ValueError(
                        "Expected call.http_request to be not None, but got None"
                    )
                request_body = call.http_request.body.json()
                cost = self.cost_estimator.calculate_cost(
                    request_body["model"],
                    call.usage.input_tokens,
                    call.usage.output_tokens,
                )
                cost_usd_per_thousand = cost["total_cost_usd"] * 1_000
                total_cost_usd_per_thousand += cost_usd_per_thousand
                all_models.add(cost["model_info"]["model_name"])

        duration_sec = f"{total_duration_ms/1000:.2f}s" if total_duration_ms else "N/A"
        return f"{total_cost_usd_per_thousand:.2f}$/1k | {duration_sec} | {', '.join(all_models)}"


class JupyterBamlMonitor(Generic[T]):
    def __init__(self, ai: T, *, summarizer=None):
        self._ai = ai
        self._summarizer = summarizer
        self._streamer: JupyterOutputBox | None = None
        self._collector: JupyterBamlCollector | None = None

    @property
    def ai(self) -> T:
        if self._collector is None:
            raise RuntimeError("JupyterTraceLLMCalls tracer has not been initialized.")
        return cast("T", self._collector.ai)

    @property
    def b(self) -> T:
        return self.ai

    async def display_calls(
        self,
        *,
        prompts: Literal["always_hide", "always_show", "show", "hide"] = "hide",
        completions: Literal["always_hide", "always_show", "show", "hide"] = "hide",
    ):
        if self._collector is None:
            raise RuntimeError("JupyterTraceLLMCalls tracer has not been initialized.")
        await self._collector.display_calls(prompts=prompts, completions=completions)

    async def display_session(
        self,
        name: str,
        *,
        prompts: Literal["always_hide", "hide"] = "hide",
        completions: Literal["always_hide", "hide"] = "hide",
        show_everything: bool = False,
    ):
        if self._collector is None:
            raise RuntimeError("JupyterTraceLLMCalls tracer has not been initialized.")
        await self._collector.display_session(
            name,
            prompts=prompts,
            completions=completions,
            show_everything=show_everything,
        )

    def __enter__(self):
        self._streamer = JupyterOutputBox(clear_after_finish=True)
        self._streamer.__enter__()
        self._collector = JupyterBamlCollector(
            self._ai,
            stream_callback=self._streamer.display(),
            intent_summarizer=self._summarizer,
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._streamer is not None:
            self._streamer.__exit__(exc_type, exc_val, exc_tb)
            self._streamer = None



from baml_agents.jupyter._export_notebook_to_py import export_notebook_to_py
from baml_agents.jupyter._jupyter_baml import (
    JupyterBamlCollector,
    JupyterBamlMonitor,
    JupyterOutputBox,
)
from baml_agents.jupyter._jupyter_chat_widget import ChatMessage, JupyterChatWidget

__all__ = [
    "ChatMessage",
    "JupyterBamlCollector",
    "JupyterBamlMonitor",
    "JupyterChatWidget",
    "JupyterOutputBox",
    "export_notebook_to_py",
]



"""
Jupyter Chat Widget.

Source: https://github.com/darinkist/MediumArticle_InteractiveChatGPTSessionsInJupyterNotebook

MIT License

Copyright (c) 2023 darinkist

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import datetime
import traceback
from collections.abc import Awaitable, Callable
from typing import Literal

from IPython.display import HTML, display
from ipywidgets import widgets
from pydantic import BaseModel, ConfigDict


class ChatMessage(BaseModel):
    content: str
    role: Literal["user", "assistant"]
    model_config = ConfigDict(frozen=True)


class JupyterChatWidget:
    def __init__(
        self,
        message_callback: (
            Callable[[list[ChatMessage]], Awaitable[list[ChatMessage]]] | None
        ),
        *,
        chat_history: list[ChatMessage] | None = None,
        user_label: str = "User",
        assistant_label: str = "Agent",
    ):
        self._chat_messages: list[ChatMessage] = []

        self._setup_css()
        self.output = widgets.Output()

        # Change to TextArea which often has better copy-paste support
        self.in_text = widgets.Textarea(
            placeholder="Type your message here...",
            layout=widgets.Layout(width="100%", height="60px"),
        )
        self.in_text.continuous_update = False

        # Add a button for sending messages as an alternative input method
        self.send_button = widgets.Button(
            description="Send", button_style="primary", icon="paper-plane"
        )
        self.send_button.on_click(self._handle_button_click)

        # Set up keyboard event handling
        self._setup_keyboard_handling()

        # Simple text loading indicator
        self.loading_bar = widgets.HTML("Loading...")
        self.loading_bar.layout.display = "none"

        self.message_callback = message_callback
        self.user_label = user_label
        self.assistant_label = assistant_label

        # Display initial chat history if provided
        if chat_history:
            for item in chat_history:
                self._add_message(content=item.content, role=item.role)

    async def run(self, initial_message: str | None = None):
        """
        Display the widget and optionally send an initial message.
        This method is awaitable and returns immediately after optional initial message is sent.
        """
        self.display()
        if initial_message and self.message_callback:
            await self._send_message(initial_message)

    async def _execute_with_loading(self, coroutine_func, *args, **kwargs):
        self.loading_bar.layout.display = "block"
        try:
            return await coroutine_func(*args, **kwargs)
        except Exception:  # noqa: BLE001
            traceback.print_exc()
            return None
        finally:
            self.loading_bar.layout.display = "none"

    def _add_message(self, content, role):
        """Add a message to the chat history."""
        self._chat_messages.append(ChatMessage(content=content, role=role))
        self._display_message(content, role)

    def _setup_keyboard_handling(self):
        # JavaScript to capture Shift+Enter to send messages
        js_code = """
        function captureEnter(event) {
            if (event.key === 'Enter' && !event.shiftKey) {
                event.preventDefault();
                document.getElementById('send-button').click();
                return false;
            }
            return true;
        }

        setTimeout(function() {
            var textareas = document.getElementsByClassName('jupyter-widgets-output-area')[0]
                            .getElementsByTagName('textarea');
            if (textareas.length > 0) {
                var textarea = textareas[textareas.length-1];
                textarea.addEventListener('keydown', captureEnter);
            }
        }, 1000);
        """

        display(
            HTML(
                f"""
        <script>
        {js_code}
        </script>
        """
            )
        )

    def _setup_css(self):
        display(
            HTML(
                """
        <link rel="stylesheet"
              href="https://cdn.jsdelivr.net/npm/bootstrap@4.5.3/dist/css/bootstrap.min.css"
              integrity="sha384-TX8t27EcRE3e/ihU7zmQxVncDAy5uIKz4rEkgIXeMed4M0jlfIDPvg6uqKI2xXr2"
              crossorigin="anonymous">
        <style>
            body{margin-top:20px;}

            .chat-message-left,
            .chat-message-right {
                display: flex;
                flex-shrink: 0
            }

            .chat-message-left {
                margin-right: auto
            }

            .chat-message-right {
                flex-direction: row-reverse;
                margin-left: auto
            }

            .time-stamp {
                margin-left: 10px;
                margin-right: 10px;
                align-self: flex-end;
            }

            /* Make the textarea more responsive to copy/paste */
            .jupyter-widgets textarea {
                font-family: inherit;
                line-height: normal;
                resize: none;
                padding: 8px;
            }
        </style>
        """
            )
        )

    def display(self):
        display(
            widgets.HBox(
                [self.output],
                layout=widgets.Layout(
                    width="100%",
                    max_height="500px",
                    display="inline-flex",
                    flex_flow="column-reverse",
                ),
            )
        )

        # Modify the input area to use a button + textarea
        display(
            widgets.Box(
                children=[self.loading_bar, self.in_text, self.send_button],
                layout=widgets.Layout(
                    display="flex", flex_flow="row", align_items="center", width="100%"
                ),
            )
        )

        # Assign ID to the send button for JavaScript to find it
        display(
            HTML(
                """
        <script>
        document.querySelector(".jupyter-button:last-child").id = "send-button";
        </script>
        """
            )
        )

    def _handle_button_click(self, _):
        """Handle clicks on the send button."""
        return asyncio.create_task(self._send_message(self.in_text.value))

    async def _send_message(self, text):
        if not text or text.strip() == "":
            return

        question = text.strip()

        # Clear the input
        self.in_text.value = ""

        # Display user message in the UI
        self._add_message(content=question, role="user")

        # Validate callback is a coroutine function
        if not asyncio.iscoroutinefunction(self.message_callback):
            raise ValueError("message_callback must be a coroutine function")

        # Execute callback with loading indicator
        response = await self._execute_with_loading(
            self.message_callback, self._chat_messages[:]
        )

        # Display the assistant response
        if response:
            for item in response:
                self._add_message(item.content, item.role)

        # Focus back to the input textarea
        display(
            HTML(
                """
        <script>
        document.querySelector(".jupyter-widgets textarea:last-of-type").focus();
        </script>
        """
            )
        )

    def _display_message(self, content, role):
        # No longer maintaining internal chat history
        content_formatted = content.replace("$", r"\$")

        # Determine if message should be displayed as user or non-user (assistant)
        is_user = role == "user"

        # Set alignment based on user status
        message_class = "chat-message-left" if is_user else "chat-message-right"

        # Display the actual role in the UI
        display_label = self.user_label if is_user else (self.assistant_label)

        timestamp = datetime.datetime.now(datetime.UTC).strftime("%H:%M:%S")
        html = (
            f'<div class="{message_class} pb-4">'
            '<div class="flex-shrink-1 bg-light rounded py-2 px-3 ml-3">'
            f'<div class="font-weight-bold mb-1">{display_label}</div>{content_formatted}</div>'
            f'<div class="text-muted small time-stamp">{timestamp}</div></div>'
        )
        self.output.append_display_data(HTML(html))

    def _clear_chat(self):
        self.output.clear_output()




from abc import ABC, abstractmethod
from typing import Any, final

from baml_agents._baml_client_proxy._hooks._base_hook import (
    BaseBamlHookAsync,
    BaseBamlHookContext,
    BaseBamlHookSync,
)


@final
class OnBeforeCallHookContext(BaseBamlHookContext):
    pass


class OnBeforeCallHookAsync(BaseBamlHookAsync, ABC):
    @abstractmethod
    async def on_before_call(
        self,
        *,
        ctx: OnBeforeCallHookContext,
        params: dict[str, Any],
    ) -> None:
        pass


class OnBeforeCallHookSync(BaseBamlHookSync, ABC):
    @abstractmethod
    def on_before_call(
        self,
        *,
        ctx: OnBeforeCallHookContext,
        params: dict[str, Any],
    ) -> None:
        pass



import os
from typing import TypeVar

from pydantic import BaseModel

from ._with_baml_client import with_baml_client

T = TypeVar("T")


class BamlModelConfig(BaseModel):
    provider: str = "openai"
    api_key: str | None = None
    api_key_env_var: str = "OPENAI_API_KEY"
    base_url: str | None = None
    base_url_env_var: str = "OPENAI_API_BASE"


def with_model(
    b: T,
    model: str,
    *,
    config: BamlModelConfig | None = None,
) -> T:
    config = config or BamlModelConfig()
    api_key = config.api_key or os.environ.get(config.api_key_env_var)
    base_url = config.base_url or os.environ.get(config.base_url_env_var)

    options = {
        "model": model,
        "api_key": api_key,
    }
    if base_url:
        options["base_url"] = base_url

    return with_baml_client(b, provider=config.provider, options=options)  # type: ignore



import os
import pathlib
import sys
from collections import defaultdict
from collections.abc import Mapping
from dataclasses import dataclass

from loguru import logger

from baml_agents._project_utils._get_root_path import get_root_path


class _Style:
    BLACK = "black"
    BOLD = "bold"
    BLUE = "blue"
    DIM = "dim"
    CYAN = "cyan"
    NORMAL = "normal"
    GREEN = "green"
    ITALIC = "italic"
    MAGENTA = "magenta"
    UNDERLINE = "underline"
    RED = "red"
    STRIKE = "strike"
    WHITE = "white"
    REVERSE = "reverse"
    YELLOW = "yellow"


@dataclass(frozen=True)
class LogColorConfig:
    text_color: str = _Style.GREEN
    key_color: str = _Style.CYAN
    value_color: str = _Style.BLUE
    comment_color: str = _Style.MAGENTA
    keyword_color: str = _Style.GREEN


class _LogFormatter:
    """Unified formatter for both HTML and terminal log output."""

    def __init__(
        self,
        root_path,
        working_directory,
        color_config: LogColorConfig,
    ):
        self.root_path = root_path
        self.working_directory = working_directory
        self._color_config = color_config

    def filter_extras(self, record):
        filtered_extras = {}
        for key, value in record["extra"].items():
            if value is None:
                continue
            if isinstance(value, Mapping):
                formatted_value = " ".join(f"{k}={v}" for k, v in value.items())
                filtered_extras[key] = formatted_value
            elif isinstance(value, float) and value.is_integer():
                filtered_extras[key] = int(value)
            else:
                filtered_extras[key] = value
        return filtered_extras

    def format_extras(self, extras):
        if not extras:
            return ""
        items = []
        for key, value in extras.items():
            key_span = f"<{self._color_config.key_color}>{key}</{self._color_config.key_color}>"
            value_span = f"<{self._color_config.value_color}>{value}</{self._color_config.value_color}>"
            items.append(f"{key_span}={value_span}")
        return " ".join(items).replace("{", "{{").replace("}", "}}").replace("\n", "")

    def compute_filepath(self, record):
        """Compute the relative filepath for the log record."""
        filepath_root = pathlib.Path(
            f'{self.root_path}/{record["name"].replace(".", "/")}.py'
        )
        filepath_cwd = pathlib.Path(
            f'{self.working_directory}/{record["name"].replace(".", "/")}.py'
        )

        if filepath_root.exists():
            filepath = filepath_root
        elif filepath_cwd.exists():
            filepath = filepath_cwd
        else:
            return record["name"]

        relative_path = os.path.relpath(
            filepath,
            self.working_directory,
        )
        return f'./{relative_path}:{record["line"]}'

    def __call__(self, record):
        filtered_extras = self.filter_extras(record)
        formatted_extras = self.format_extras(filtered_extras)
        cc = self._color_config
        parts = [
            f"<{cc.text_color}>{{time:HH:mm:ss}}</{cc.text_color}> ",
            f"<{cc.keyword_color}><level>{{level:<8}}</level></{cc.keyword_color}> ",
            "{message}",
        ]
        if formatted_extras:
            parts.append(f" {formatted_extras}")
        parts.append(
            f" <{cc.keyword_color}><level>FROM</level></{cc.keyword_color}> <{cc.comment_color}>{{function}} <{cc.keyword_color}><level>IN</level></{cc.keyword_color}> {self.compute_filepath(record)}</{cc.comment_color}>"
        )
        return f'{"".join(parts)}\n'


def _group_per_package(per_package):
    grouped_per_package = {}
    if per_package:
        grouped = defaultdict(list)
        for package, value in per_package.items():
            grouped[value].append(package)
        # Sort the module lists for each value
        grouped_per_package = {
            value: sorted(modules, reverse=True)
            for value, modules in sorted(
                grouped.items(), key=lambda item: item[0], reverse=True
            )
        }
    return grouped_per_package


def _add_loggers(level, per_package, formatter):
    # The goal is to ensure that for each (level, modules) group, only those modules get that level,
    # and all other log records (not matching any module in any group) get the default level.
    # This is a "1 vs all" filter per group.

    grouped_per_package = _group_per_package(per_package)

    # Track all modules in all groups for the "all others" filter
    all_grouped_modules = set()
    for modules in grouped_per_package.values():
        all_grouped_modules.update(modules)

    # Add a logger for each group: only log records whose name starts with one of the group's modules
    for value, modules in grouped_per_package.items():
        module_prefixes = tuple(modules)
        logger.add(
            sys.stdout,
            filter=lambda record, module_prefixes=module_prefixes: record[
                "name"
            ].startswith(  # type: ignore
                module_prefixes
            ),
            format=formatter,
            level=value,
        )

    # Add a logger for all other modules (not in any group)
    logger.add(
        sys.stdout,
        filter=lambda record: not record["name"].startswith(tuple(all_grouped_modules)),  # type: ignore
        format=formatter,
        level=level,
    )


def init_logging(
    *,
    level: str | None = None,
    per_package: dict[str, str] | None = None,
    root_path=None,
    working_dir=None,
    color_config: LogColorConfig | None = None,
):
    level = level or "INFO"
    root_path = root_path or get_root_path()
    color_config = color_config or LogColorConfig()
    working_dir = working_dir or pathlib.Path.cwd()
    formatter = _LogFormatter(root_path, working_dir, color_config=color_config)
    logger.remove()
    _add_loggers(level, per_package, formatter)




import json
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar, runtime_checkable

from ._model import BamlClassModel, BamlEnumModel

if TYPE_CHECKING:
    from baml_py.type_builder import TypeBuilder

T = TypeVar("T", bound="TypeBuilder")


class BaseSchemaConverter:
    def __init__(self, schema: str | Mapping[str, Any]):
        if isinstance(schema, str):
            try:
                self._root_schema: dict[str, Any] = json.loads(schema)
            except json.JSONDecodeError as e:
                raise ValueError(f"Invalid JSON schema provided: {e}") from e
        elif isinstance(schema, Mapping):
            self._root_schema = dict(schema)
        else:
            raise TypeError(
                f"Schema must be a JSON string or a dictionary, got {type(schema)}",
            )


class AbstractMcpSchemaToBamlModelConverter(ABC, BaseSchemaConverter):
    @abstractmethod
    def convert_tools(self) -> list[BamlClassModel | BamlEnumModel]:
        pass


class AbstractJsonSchemaToBamlModelConverter(ABC, BaseSchemaConverter):
    def __init__(self, schema: str | Mapping[str, Any], class_name: str):
        super().__init__(schema)
        self._class_name = class_name

    @abstractmethod
    def convert(self) -> list[BamlClassModel | BamlEnumModel]:
        pass


class BamlModelConverter:
    def __init__(self, baml_models: list[BamlClassModel | BamlEnumModel]):
        # Check for duplicate model names
        names = [model.name for model in baml_models]
        duplicates = {name for name in names if names.count(name) > 1}
        if duplicates:
            raise ValueError(
                f"Duplicate model names found: {', '.join(sorted(duplicates))}"
            )

        self._baml_models = baml_models


class BamlSourceGenerator(ABC, Generic[T], BamlModelConverter):
    """Abstract base for generating BAML source code."""

    @abstractmethod
    def generate(self) -> str:
        """
        Generate BAML source code.

        :return: BAML source code as a string.
        """


class BamlTypeBuilderConfigurer(ABC, Generic[T], BamlModelConverter):
    """Abstract base for configuring BAML TypeBuilder objects by adding types/classes."""

    @abstractmethod
    def configure(self, tb: T):
        """
        Add types/classes into the provided TypeBuilder.

        :param tb: An instance of TypeBuilder to be configured.
        """


@runtime_checkable
class ArgsClassCallback(Protocol):
    def __call__(
        *,
        name: str,
        schema: dict[str, Any],
    ) -> str: ...


@runtime_checkable
class ToolClassCallback(Protocol):
    def __call__(
        *,
        name: str,
        schema: dict[str, Any],
    ) -> str: ...


@runtime_checkable
class ToolNameCallback(Protocol):
    def __call__(
        *,
        name: str,
        schema: dict[str, Any],
    ) -> str: ...


@runtime_checkable
class PropNameCallback(Protocol):
    def __call__(
        *,
        name: str,
        prop_schema: dict[str, Any],
        class_schema: dict[str, Any],
        root_class_schema: dict[str, Any],
    ) -> str: ...


@runtime_checkable
class DescCallback(Protocol):
    def __call__(
        *,
        description: str | None,
        root: bool,
        prop_schema: dict[str, Any],
        class_schema: dict[str, Any],
        root_class_schema: dict[str, Any],
    ) -> str | None: ...


@runtime_checkable
class AliasCallback(Protocol):
    def __call__(
        *,
        name: str,
        root: bool,
        prop_schema: dict[str, Any],
        class_schema: dict[str, Any],
        root_class_schema: dict[str, Any],
    ) -> str | None: ...




from collections.abc import Iterable


def sole(collection: Iterable):
    """
    Return the sole (one and only) element from the collection.

    Args:
        collection: An iterable with exactly one element

    Returns:
        The single element

    Raises:
        ValueError: If collection doesn't contain exactly one element

    """
    iterator = iter(collection)
    try:
        first_item = next(iterator)
    except StopIteration as e:
        raise ValueError("Expected exactly one element, found 0") from e
    try:
        next(iterator)
        raise ValueError("Expected exactly one element, found more than 1")
    except StopIteration:
        return first_item
